def ui_raw(options):
    
    import tkinter as tk
    import numpy as np

    win = tk.Tk()
    win.configure(background='white')
    win.title('Manual component classification. Choose and click OK to continue.')

    if options['figsize'] == 'small':
        sz = [700, 560]
    elif options['figsize'] == 'medium':
        sz = [900, 600]
    elif options['figsize'] == 'large':
        sz = [1200, 900]

    # center window horizontally and vertically
    sw = win.winfo_screenwidth()
    xpos = int(np.ceil((sw-sz[0])/2))

    sh = win.winfo_screenheight()
    ypos = int(np.ceil((sh-sz[1])/2))

    win.geometry(f'{sz[0]}x{sz[1]}+{xpos}+{ypos}')

    return win


def ui_check(self):

    import numpy as np
    import tkinter as tk
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
    import matplotlib.pyplot as plt

    # extract data from instance
    ch_idx = [  i for i, chan in enumerate(self.epochs.ch_names) 
                if chan in self.options['chanpicks']]

    post= np.mean(self.post, 2)
    pre = np.mean(self.epochs._data[:,ch_idx, :], 0)
    rel_time = self.epochs.times*1000

    win = ui_raw(self.options)
    win.configure(background='white')
    win.title('Are you satisfied with the correction?')

    # plot pre-/post
    fig, (sp1, sp2) = plt.subplots(2)

    sp1.plot(rel_time, pre.T, 'k', linewidth=0.5)
    sp1.title.set_text('Pre')
    sp1.set_xlabel('time [ms]')
    sp1.set_ylabel('amplitude [\muV]') 

    sp2.plot(rel_time, post.T, 'k', linewidth=0.5)
    sp2.title.set_text('Post')
    sp2.set_xlabel('time [ms]')
    sp2.set_ylabel('amplitude [\muV]') 

    # make ylimits equal in both plots
    sp2.set_ylim(sp1.get_ylim())

    canvas = FigureCanvasTkAgg(fig)
    widget = canvas.get_tk_widget()
    widget.pack(fill=tk.BOTH, expand=tk.YES, side='bottom')
    canvas.draw()

    buttonframe = tk.Frame(win, padx=10, pady=10, bg='white')
    buttonframe.pack(side='top')

    def on_yes():
        global redo
        redo = False
        win.destroy()

    def on_no():
        global redo
        redo = True
        win.destroy()

    yesbutton = tk.Button(buttonframe, bg='white', text='Yes', command=on_yes, padx=20)
    yesbutton.pack(side='left')

    nobutton = tk.Button(buttonframe, bg='white', text='No', command=on_no, padx=20)
    nobutton.pack(side='left')

    win.mainloop()

    return redo


def ui_redo():

    import tkinter as tk

    win = tk.Tk()
    win.configure(background='white')
    win.title('Redo with new settings?')

    buttonframe = tk.Frame(win, padx=10, pady=10, bg='white')
    buttonframe.pack(side='top')

    def on_yes():
        global newsettings
        newsettings = True
        win.destroy()

    def on_no():
        global newsettings
        newsettings = False
        win.destroy()

    yesbutton = tk.Button(buttonframe, bg='white', text='Yes, redo.', command=on_yes, padx=20)
    yesbutton.pack(side='left')

    nobutton = tk.Button(buttonframe, bg='white', text='No, only re-check components.', command=on_no, padx=20)
    nobutton.pack(side='left')

    win.mainloop()  

    return newsettings



def ui_select(inst):

    import tkinter as tk
    from mne.viz import plot_topomap
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
    import matplotlib.pyplot as plt
    from scipy.stats import zscore
    from mne.channels import read_layout
    import numpy as np


    rel_time = inst.epochs.times*1000
    ch_names = [chan for chan in inst.epochs.ch_names 
                if chan in inst.options['chanpicks']]

    # color code for plotting (red = suspected artifact)
    clrs = np.where(inst.compclass == 1, 'b', 'r')


    for compnum in range(inst.options['comps']):

        win = ui_raw(inst.options)
        fig, ((sp1, sp2), (sp3, sp4)) = plt.subplots(2, 2)

        temp = inst.S[compnum, :, :]

        # plot time series of component
        sp1.plot(rel_time, np.mean(temp, 1), clrs[compnum])
        sp1.title.set_text('Component time series across channels')
        sp1.set_xlabel('Time [ms]')
        sp1.set_ylabel('Amplitude [a.u.]')

        # plot topographical plot
        # adjust channel positions from default mne layout, 
        # original from mkeute (github), plot topographical plot with mne
        layout = read_layout("EEG1005")
        pos = (np.asanyarray([layout.pos[layout.names.index(ch)] for ch in ch_names])[:, 0:2]- 0.5) / 5
        plot_topomap(inst.A[:,compnum].T, pos, names=ch_names, show_names=True, axes=sp2, show=False)
        sp2.title.set_text('Topographical map')
        
        # plot spectral information
        freq = np.arange(   inst.options['plotfreqx'][0], 
                            inst.options['plotfreqx'][1]+ 0.5, 0.5)

        if inst.options['freqscale'] == 'raw':
            sp3.plot(freq, inst.fftbins[compnum, :].T)
        elif inst.options['freqscale'] == 'log':
            sp3.plot(freq, np.log(inst.fftbins[compnum, :].T))
        elif inst.options['freqscale'] == 'log10':
            sp3.plot(freq, np.log10(inst.fftbins[compnum, :].T))
        elif inst.options['freqscale'] == 'db':
            sp3.plot(freq, 10*np.log10(inst.fftbins[compnum, :].T))

        sp3.title.set_text('Power spectrum')
        sp3.set_xlabel('Frequency [Hz]')
        sp3.set_ylabel('Power [muV^2/Hz]')

        # plot time course matrix
        tp1 = np.argmin(np.abs(rel_time - inst.options['plottimex'][0]))
        tp2 = np.argmin(np.abs(rel_time - inst.options['plottimex'][1]))
        sp4.matshow(np.transpose(zscore(temp[tp1:tp2, :])), extent=[0, 1, 0, 0.8], interpolation=None, cmap='RdBu')
        sp4.title.set_text('Time course matrix')
        sp4.set_xlabel('Time [ms]')
        sp4.set_ylabel('Trials')

        plt.tight_layout()
        canvas = FigureCanvasTkAgg(fig)
        canvas.get_tk_widget().pack(fill=tk.BOTH, expand=tk.YES, side='bottom')
        canvas.draw()

        def on_closing():

            choice = clicked.get()
            inst.compclass[compnum] = int(art_types.index(choice)) + 1
            win.destroy()
        
        buttonframe = tk.Frame(win, padx=10, pady=10, bg='white')
        buttonframe.pack(side='right')

        okbutton = tk.Button(buttonframe, bg='white', text='OK', command=on_closing)
        okbutton.pack(side='right')

        # make a menu for choosing whether component should be rejected or not. Default should be based on threshold.
        clicked = tk.StringVar()

        art_types = ["neural", "tms muscle artifact", 
                    "eyeblink", "lateral eye movement", 
                    "persistent muscle activity", "electrode noise"]

        # use the classification number to index the list with the different artifact types
        # and set as default value in dropdown menu
        clicked.set(art_types[int(inst.compclass[compnum])-1])

        menu = tk.OptionMenu(buttonframe, clicked, *art_types)
        menu.configure(background='white', activebackground='white')
        menu.pack(side='right')

        win.mainloop()




def textfield (win, default, rownr, colnr, sticky='w'):

    import tkinter as tk

    inputvar = tk.StringVar()

    if type(default) == list and len(default) == 2:
        inputvar.set('[' + str(default[0]) + ', ' + str(default[1]) + ']')
    else:
        inputvar.set(default)

    txtfld = tk.Entry(win, text=inputvar)
    txtfld.configure(background='white', width=10)
    txtfld.grid(row=rownr, column=colnr, sticky=sticky)

    return inputvar



def toggle(frame, default, row, column):

    import tkinter as tk

    var = tk.BooleanVar()
    if default =='on':
        var.set(True)
    else:
        var.set(False)

    r_on = tk.Radiobutton(frame, text='on', background='white', variable=var, value=True)
    r_on.grid(row=row, column=column, sticky='w')

    r_off = tk.Radiobutton(frame, text='off', background='white', variable=var, value=False)
    r_off.grid(row=row, column=column, sticky='e')

    return var



def choicemenu(frame, default, choices, row, column):

    import tkinter as tk

    var =  tk.StringVar(value=default)
    menu = tk.OptionMenu(frame, var, *choices)
    menu.configure(background='white', activebackground='white', width=10)
    menu.grid(row=row, column=column, sticky='w')

    return var



def ui_input(options):

    import tkinter as tk
    import ast
    import numpy as np

    win = ui_raw(options)
    win.title('TESA - Please specify settings.')
    ttl_font = ('Helvetica', 12, 'bold')
    framepadx = 20
    framepady = 20

    # Frames
    frame_fig = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_feedb = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_tmsmuscle = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    
    frame_ica = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_eye = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_lat = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_muscle = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)
    frame_elnoise = tk.Frame(win, bg='white', padx=framepadx, pady=framepady)

    frames = [frame_fig, frame_feedb, frame_tmsmuscle, frame_ica, 
                frame_eye, frame_lat, frame_muscle, frame_elnoise]
    titles = ['Figure Settings', 'Feedback Settings', 'TMS muscle artifacts', 'FastICA Settings', 
                'Eyeblinks', 'Lateral eye movement', 'Persistent muscle activity', 'Electrode noise']

    # Frame titles
    for i, ttl in enumerate(titles):
        tk.Label(frames[i], text=ttl, background='white', font=ttl_font).grid(row=0, column=0, sticky='w')


    # FIGURE SETTINGS
    tk.Label(frame_fig, text='Figure Size: ', background='white').grid(row=1, column=0, sticky='w')
    figsize = choicemenu(frame_fig, options['figsize'], ('small', 'medium', 'large'), 1, 1)

    tk.Label(frame_fig, text='Time window [ms]: ', background='white').grid(row=2, column=0, sticky='w')
    plottimex = textfield(frame_fig, options['plottimex'], 2, 1, 'w')

    tk.Label(frame_fig, text='Frequencies to plot [Hz]: ', background='white').grid(row=3, column=0, sticky='w')
    plotfreqx = textfield(frame_fig, options['plotfreqx'], 3, 1, 'w')

    tk.Label(frame_fig, text='Frequency scaling: ', background='white').grid(row=4, column=0, sticky='w')
    freqscale = choicemenu(frame_fig, options['freqscale'], ('raw', 'log', 'log10', 'db'), 4, 1)

    frame_fig.grid(row=0, column=0, sticky='nw')
    frame_fig.grid_columnconfigure(1, minsize=100)



    # FEEDBACK SETTINGS
    tk.Label(frame_feedb, text='Number of components: ', background='white').grid(row=1, column=0, sticky='w')
    comps = textfield(frame_feedb, options['comps'], 1, 1, 'w')

    # Manual component check
    tk.Label(frame_feedb, text='Manual component check: ', background='white').grid(row=2, column=0, sticky='w')
    compcheck = toggle(frame_feedb, options['compcheck'], 2, 1)

    # Feedback on thresholds
    tk.Label(frame_feedb, text='Feedback on thresholds: ', background='white').grid(row=3, column=0, sticky='w')
    threshfeedback = toggle(frame_feedb, options['threshfeedback'], 3, 1)

    # Remove components
    tk.Label(frame_feedb, text='Remove selected components: ', background='white').grid(row=4, column=0, sticky='w')
    remove = toggle(frame_feedb, options['remove'], 4, 1)

    frame_feedb.grid(row=0, column=1, sticky='nw')
    frame_feedb.grid_columnconfigure(1, minsize=100)



    # TMS-EVOKED MUSCLE ARTIFACTS
    tmsmuscle = toggle(frame_tmsmuscle, options['tmsmuscle'], 1, 1)

    tk.Label(frame_tmsmuscle, text='TMS muscle threshold: ', background='white').grid(row=2, column=0, sticky='w')
    tmsmusclethresh = textfield(frame_tmsmuscle, options['tmsmusclethresh'], 2, 1, 'w')

    tk.Label(frame_tmsmuscle, text='TMS muscle window: ', background='white').grid(row=3, column=0, sticky='w')
    tmsmusclewin = textfield(frame_tmsmuscle, options['tmsmusclewin'], 3, 1, 'w')

    frame_tmsmuscle.grid(row=1, column=0, sticky='nw')
    frame_tmsmuscle.grid_columnconfigure(1, minsize=100)



    # ICA SETTINGS
    tk.Label(frame_ica, text='Approach: ', background='white').grid(row=1, column=0, sticky='w')
    approach = choicemenu(frame_ica, options['approach'], ('parallel', 'deflation'), 1, 1)

    tk.Label(frame_ica, text='Contrast function (g): ', background='white').grid(row=2, column=0, sticky='w')
    g = choicemenu(frame_ica, options['g'], ('logcosh', 'exp', 'cube'), 2, 1)

    frame_ica.grid(row=1, column=1, sticky='nw')
    frame_ica.grid_columnconfigure(1, minsize=100)



    # EYEBLINKS
    blink = toggle(frame_eye, options['blink'], 1, 1)

    tk.Label(frame_eye, text='Blink threshold: ', background='white').grid(row=2, column=0, sticky='w')
    blinkthresh = textfield(frame_eye, options['blinkthresh'], 2, 1, 'nsew')

    tk.Label(frame_eye, text='Electrodes used: ', background='white').grid(row=3, column=0, sticky='w')
    blinkelecs = textfield(frame_eye, options['blinkelecs'], 3, 1, 'nsew')

    frame_eye.grid(row=2, column=0, sticky='nw')
    frame_eye.grid_columnconfigure(1, minsize=100)



    # LATERAL EYE MOVEMENTS
    move = toggle(frame_lat, options['move'], 1, 1)

    tk.Label(frame_lat, text='Lateral movement threshold: ', background='white').grid(row=2, column=0, sticky='w')
    movethresh = textfield(frame_lat, options['movethresh'], 2, 1, 'nsew')

    tk.Label(frame_lat, text='Electrodes used: ', background='white').grid(row=3, column=0, sticky='w')
    moveelecs = textfield(frame_lat, options['moveelecs'], 3, 1, 'nsew')

    frame_lat.grid(row=2, column=1, sticky='nw')
    frame_lat.grid_columnconfigure(1, minsize=100)



    # PERSISTENT MUSCLE ACTIVITY
    muscle = toggle(frame_muscle, options['muscle'], 1, 1)

    tk.Label(frame_muscle, text='Muscle activity threshold: ', background='white').grid(row=2, column=0, sticky='w')
    musclethresh = textfield(frame_muscle, options['musclethresh'], 2, 1, 'nsew')

    tk.Label(frame_muscle, text='Frequency range to include: ', background='white').grid(row=3, column=0, sticky='w')
    musclefreqin = textfield(frame_muscle, options['musclefreqin'], 3, 1, 'nsew')

    tk.Label(frame_muscle, text='Frequency range to exclude: ', background='white').grid(row=4, column=0, sticky='w')
    musclefreqex = textfield(frame_muscle, options['musclefreqex'], 4, 1, 'nsew')

    frame_muscle.grid(row=3, column=0, sticky='nw')
    frame_muscle.grid_columnconfigure(1, minsize=100)



    # ELECTRODE NOISE
    elecnoise = toggle(frame_elnoise, options['elecnoise'], 1, 1)

    tk.Label(frame_elnoise, text='Electrode noise threshold: ', background='white').grid(row=2, column=0, sticky='w')
    elecnoisethresh = textfield(frame_elnoise, options['elecnoisethresh'], 2, 1, 'nsew')

    frame_elnoise.grid(row=3, column=1, sticky='nw')
    frame_elnoise.grid_columnconfigure(1, minsize=100)


    def on_ok():

        global choices
        choices = {}

        choices['figsize'] = figsize.get()
        choices['plottimex'] = ast.literal_eval(plottimex.get())
        choices['plotfreqx'] = ast.literal_eval(plotfreqx.get())
        choices['freqscale'] = freqscale.get()

        choices['tmsmuscle'] = tmsmuscle.get()

        if comps.get().strip().isnumeric():
                choices['comps'] = ast.literal_eval(comps.get())

        choices['compcheck'] = compcheck.get()
        choices['threshfeedback'] = threshfeedback.get()
        choices['remove'] = remove.get()

        if tmsmuscle.get() == True:
            if tmsmusclethresh.get().strip().isnumeric():
                choices['tmsmusclethresh'] = ast.literal_eval(tmsmusclethresh.get())
            else:
                Warning('Input for the tms muscle threshold could not be evaluated. No threshold set.')

            choices['tmsmusclewin'] = ast.literal_eval(tmsmusclewin.get())

        choices['approach'] = approach.get()
        choices['g'] = g.get()

        choices['blink'] = blink.get()

        if blink.get() == True:
            if blinkthresh.get().strip().isnumeric():
                choices['blinkthresh'] = ast.literal_eval(blinkthresh.get())
            else:
                Warning('Input for the eyeblink threshold could not be evaluated. No threshold set.')

            choices['blinkelecs'] = [str.strip("[] ") for str in blinkelecs.get().split(",")]

        choices['move'] = move.get()

        if move.get() == True:
            if movethresh.get().strip().isnumeric():
                choices['movethresh'] = ast.literal_eval(movethresh.get())
            else:
                Warning('Input for the lateral eye movement threshold could not be evaluated. No threshold set.')

            choices['moveelecs'] = [str.strip("[] ") for str in moveelecs.get().split(",")]

        choices['muscle'] = muscle.get()
        
        if muscle.get() == True:
            if musclethresh.get().strip().isnumeric():
                choices['musclethresh'] = ast.literal_eval(musclethresh.get())
            else:
                Warning('Input for the persistent muscle activity threshold could not be evaluated. No threshold set.')

            if len(musclefreqin.get()) > 0:
                choices['musclefreqin'] = ast.literal_eval(musclefreqin.get())
            if len(musclefreqex.get()) > 0:
                choices['musclefreqex'] = ast.literal_eval(musclefreqex.get())

        choices['elecnoise'] = elecnoise.get()

        if elecnoise.get() == True:
            if elecnoisethresh.get().strip().isnumeric():
                choices['elecnoisethresh'] = float(elecnoisethresh.get())
            else:
                Warning('Input for the electrode noise threshold could not be evaluated. No threshold set.')

        win.destroy()

    def on_cancel():
        win.destroy()

    frame_button = tk.Frame(win, padx=10, pady=10, bg='white')
    frame_button.grid(row=4, column=1)

    cancelbutton = tk.Button(frame_button, bg='white', text='Cancel Changes', command=on_cancel)
    cancelbutton.grid(row=0, column=0, sticky='ne')
    okbutton = tk.Button(frame_button, bg='white', text='Ok', command=on_ok)
    okbutton.grid(row=0, column=1, sticky='ne')

    win.columnconfigure((0,1), weight=50)
    win.rowconfigure((0,1,2,3), weight=50)

    win.mainloop()

    try:
        for key, val in choices.items():

            if val == True:
                val = 'on'
            elif val == False:
                val = 'off'

            options[key] = val
    except:
        print('\nOptions not changed.')

    return options




import TMSRepair.TMSRepair_misc as misc
import TMSRepair.TMSRepair_UIs as UIs

import numpy as np
import matplotlib
import mne

from copy import copy
from sklearn.decomposition import FastICA
from scipy.stats import zscore




class TMSepochs:
    """
    Class for repairing TMS evoked artifacts in 
    mne epochs objects based on fast ICA.


    Attributes
    ----------
    epochs : mne epochs object
        data to be repaired in mne format
    options : dict
        dictionary with the settings for the ICA. 
    rank : int
        rank of the matrix that the ICA is performed on
    orig_backend : string
        original backend at the timepoint the object is initialized. 
        The backend needs to be changed to 'Agg' for tkinter 
        and later be reset.
    S : numpy array
        ICA components time courses
    A : numpy array
        mixing matrix
    badcomp : list
        the components that were chosen to be rejected
    compclass : list
        list of length (n° components) with values ranging from 0-6 
        indicating the class of the component
    fftbins : numpy array
        spectral information of each component
    mean : numpy array
        mean for each channel that was subtracted from the data for whitening it prior to ICA, 
        that needs to be added after the inverse transform  
    perc_var : array
        percent of amplitude variance relative to other components over time 
        for each component
    post : numpy array
        data after the inverse transform, stored separately
        in order not to replace it in the epochs object right away
    transformed : bool
        whether the data in the epochs object has already been transformed 
        and the matrix rank adjusted to the component rejection


    Methods
    ----------
    mark_bad_channels:
        interactive rejection of bad channels specifically for analyzing event-related potentials.
        plots channel variance and the event related potentials 
        averaged across epochs for bad channel marking.
    replace_with_zeros(window : list):
        removes the data in each epoch 
        in the specified window [ms, e.g. [-5, 15]] and replaces it with zeros
    cubic_interpolation(window : list):
        interpolates the data in each epoch 
        in the specified window with first degree cubic interpolation
    fastica:
        performs fast ICA and sorts components after their variance over time
    compselect:
        component classification based on thresholds and/or visual inspection
    inverse_transform:
        application of the inverse transform, rejection of artifactual components,
        optionally visual check. 
        Result is saved in self.post (data in epochs not changed)
    transform_epochs_object:
        replace epochs._data with the transformed data
    reset_orig_backend:
        reset the matplotlib backend after setting it to 'Agg' for tkinter
    fit_select_transform:
        wrapper method calling the methods: fastica(), compselect(), 
        inverse_transform(), transform_epochs_object(),
        reset_orig_backend()

    """

    # define default options
    options = { 'manualinput':'on',
                'confirm':'on',
                'threshfeedback': 'on',
                'chanpicks':[],
                'compcheck':'on',
                'remove':'on', 
                'comps': -1, 
                'figsize': 'medium', 
                'plottimex': [-200, 300], 
                'plotfreqx': [1,100],
                'freqscale': 'log',
                
                'tmsmuscle': 'on',
                'tmsmusclethresh': 8,
                'tmsmusclewin': [11,30],

                'approach': 'parallel', 
                'g': 'logcosh', 

                'blink':'on', 
                'blinkthresh':2.5, 
                'blinkelecs':['Fp1', 'Fp2'], 

                'move':'on', 
                'movethresh':2, 
                'moveelecs':['F7', 'F8'],

                'muscle':'on', 
                'musclethresh':-0.31, 
                'musclefreqin':[],
                'musclefreqex':[48, 52],
                
                'elecnoise':'on',
                'elecnoisethresh':2}




    def __init__(self, epochs, options=None):

        self.epochs = epochs
        self.options['chanpicks'] = [self.epochs.ch_names[i] for i in self.epochs.picks]
        self.orig_backend = matplotlib.get_backend()

        # overwrite default options with user choices and check them
        if options is not None:
            self.set_options(options)
        else:
            self.set_options({})




    def set_options(self, options):

        self.options = misc.eval_param(self.options, options)  

        # if manual input desired open UI
        if self.options['manualinput'] == 'on':
            self.options = UIs.ui_input(self.options)

        # check if parameters are valid
        misc.check_param(self)



    def replace_with_zeros(self, win:list):
        
        # convert window to indices in epochs, +1 to include last timepoint
        idx1 = np.argmin(np.abs(self.epochs.times - win[0]*0.001))
        idx2 = np.argmin(np.abs(self.epochs.times - win[1]*0.001)) +1

        # set values in specified window to 0
        self.epochs._data[:, :, idx1:idx2] = 0




    def cubic_interpolation(self, win:list):

        self.epochs = misc.cubic_interpolation(self.epochs, win)




    def mark_bad_channels(self):
            
        # plotting of channel variance
        vars = np.var(self.epochs._data.T, axis=(0,2))
        badchans = misc.chan_visual_inspection(vars)

        # filtering and plotting of epochs data with marked bad channels
        # bandpass and bandstop filter data first
        epochs = copy(self.epochs)
       
        # mark the channels with high variance
        epochs.info['bads'] = [epochs.info['ch_names'][i] for i in list(badchans)]

        # interpolate, bandpass and bandstop filter data
        # this is done only on the copied epochs object within the scope of this function, 
        # only for visualization and noisy channel detection!
        epochs = misc.cubic_interpolation(epochs, [-5, 15])
        epochs.filter(0.5, 49, method='iir', verbose=0)
        epochs._data = mne.filter.notch_filter( epochs._data, epochs.info['sfreq'], 50, notch_widths=2, 
                                                phase='zero', verbose=0)

        # create a fake raw object out of the evoked object, 
        # to see which channels distort the ERP and be able to mark them
        evoked = epochs.average()
        fake_raw = misc.MNE_raw_format( evoked._data.T, 
                                        epochs.info['ch_names'], 
                                        epochs.info['sfreq'])

        fake_raw.info['bads'] = epochs.info['bads']

        fake_raw.plot(  n_channels=len(epochs.info['ch_names']), scalings='auto', 
                        title='EEG data - Choose bad channels manually', 
                        show=True,
                        bad_color='r',
                        block=True)

        # take bad channels out of the channel picks
        self.options['chanpicks'] = [   chan for chan in self.options['chanpicks'] 
                                        if chan not in fake_raw.info['bads']]

        self.epochs.info['bads'] = fake_raw.info['bads']


        print('Number of rejected channels: {}'.format(len(fake_raw.info['bads'])))




    def fastica(self):

        ch_idx = [ i for i, chan in enumerate(self.epochs.ch_names) 
                        if chan in self.options['chanpicks']]

        # reconcatenate epochs
        nevents, nchans, npnts = np.shape(self.epochs._data[:, ch_idx, :])
        data_concat = np.reshape(np.moveaxis(self.epochs._data[:, ch_idx, :], 0, 2), [nchans, -1])

        print('\nPerforming fast ICA on data using {} approach.'
                .format(self.options['approach']))

        # check whether matrix is full rank, or adjust the number of components
        # otherwise fast ICA may fail to converge because it is searching for more ICs 
        # than there are in the data
        self.rank = np.linalg.matrix_rank(data_concat)

        if self.rank < self.options['comps']:
            print('The matrix rank is {}. '. format(self.rank))
            print('Number of components adjusted accordingly.')
            self.options['comps'] = self.rank


        # run FastICA and reshape component time courses
        ica = FastICA(  n_components=self.rank, 
                        algorithm=self.options['approach'], 
                        fun=self.options['g'], 
                        max_iter=1000)

        ica = ica.fit(data_concat.T)
        icasig = ica.transform(data_concat.T)

        self.S = np.reshape(icasig.T, [-1, npnts, nevents]) # component 3d time courses
        self.A = ica.mixing_ # topographies
        self.mean = ica.mean_ # mean for the inverse transform of the whitened data

        # get variance of each component in percent relative to all components as mean over epochs
        vars = np.var(np.mean(self.S, 2), axis=1)
        self.perc_var =  vars/sum(vars)*100

        # sort components in descending order based on variance
        ixsSort = np.flip(np.argsort(self.perc_var))

        self.perc_var = self.perc_var[ixsSort]
        self.A = self.A[:, ixsSort]
        self.S = self.S[ixsSort, :,:]

        print('\nICA weights sorted by time course variance.')




    def compselect(self):

        # create zscore for each component across channels
        tempCompZ = zscore(self.A, 0)

        # tms muscle window
        if self.options['tmsmuscle'] == 'on':
            mt1 = np.argmin(np.abs(self.epochs.times*1000 - self.options['tmsmusclewin'][0]))
            mt2 = np.argmin(np.abs(self.epochs.times*1000 - self.options['tmsmusclewin'][1]))

            muscleScore = np.abs(np.mean(self.S, 2))
            winScore = np.mean(muscleScore[:, mt1:mt2], 1)
            tmsmuscleratio = winScore / np.mean(muscleScore)

        # eyeblinks
        if self.options['blink'] == 'on':
            blinkidx = [i for i, chan in enumerate(self.options['chanpicks']) if chan in self.options['blinkelecs']]
            blinkratio = np.mean(tempCompZ[blinkidx, :], 0)

        # lateral eye movements
        if self.options['move'] == 'on':
            moveidx = [i for i, chan in enumerate(self.options['chanpicks']) if chan in self.options['moveelecs']]
            moveval = np.transpose(tempCompZ[moveidx,:])

        # electrode noise
        if self.options['elecnoise'] == 'on':
            elecnoise = np.amax(np.abs(tempCompZ), axis=0)

        # get indices for frequency range to detect persistent muscle activity
        freq = np.arange(self.options['plotfreqx'][0], self.options['plotfreqx'][1]+ 0.5, 0.5)

        # if needed for muscle activity detection or component inspection, calculate frequency spectrum
        if self.options['muscle'] == 'on' or self.options['compcheck'] == 'on':

            sfreq = self.epochs.info['sfreq']
            _, L, n_epoch = np.shape(self.S)

            # find the next power of 2 from the length of Y
            NFFT = 2**np.ceil(np.log2(abs(L)))
            f = sfreq/2 * np.linspace(0, 1, int(NFFT/2+1))
            freq = np.arange(self.options['plotfreqx'][0], self.options['plotfreqx'][1]+ 0.5, 0.5)

            Y2 = np.zeros([self.options['comps'], len(freq), n_epoch])

            # get the spectral information of each component of interest
            Y = np.fft.rfft(zscore(self.S, axis=1)[:self.options['comps'],:,:], n=int(NFFT), axis=1)/L
            Yout = np.abs(Y)**2

            # create frequency bins of 0.5 Hz in width centered around whole frequencies 
            for ia, a in enumerate(freq):
                index1 = np.argmin(np.abs(f-(a-0.25)))
                index2 = np.argmin(np.abs(f-(a+0.25)))
                Y2[:, ia, :] = np.mean(Yout[:, index1:index2,:], 1)

            self.fftbins = np.mean(Y2, 2)

        
        if self.options['muscle'] == 'on':

            # frequencies to include into fit
            if len(self.options['musclefreqin']) != 0:
                fin1 = np.argmin(np.abs(freq-self.options['musclefreqin'][0]))
                fin2 = np.argmin(np.abs(freq-self.options['musclefreqin'][1]))
                freqHz = freq[fin1:fin2]
            else:
                freqHz = freq

            # frequencies to exclude from fit
            if len(self.options['musclefreqex']) != 0:
                fex1 = np.argmin(np.abs(freqHz-self.options['musclefreqex'][0]))
                fex2 = np.argmin(np.abs(freqHz-self.options['musclefreqex'][1]))
                np.delete(freqHz, slice(fex1, fex2))
            
            # get idx of frequencies in power spectrum
            musclefidx = [i for i, f in enumerate(freq) if f in freqHz]

            # polynomial fit for each component, store the slope
            freqPow = self.fftbins[:, musclefidx]
            p = np.polyfit(np.log(freqHz), np.log(freqPow).T, 1)
            muscleratio = p[0,:]

        # create output storage
        self.compclass = np.zeros([self.options['comps']])

        # select if component is artifact
        print('\nClassifying components.')

        for compnum in range(self.options['comps']):

            if self.options['tmsmuscle'] == 'on' and tmsmuscleratio[compnum] >= self.options['tmsmusclethresh']:
                self.compclass[compnum] = 2

            elif self.options['blink'] == 'on' and np.abs(blinkratio[compnum]) >= self.options['blinkthresh']:
                self.compclass[compnum] = 3

            elif self.options['move'] == 'on' and moveval[compnum, 0] >= self.options['movethresh'] and moveval[compnum, 1] <= -self.options['movethresh']  \
                or  self.options['move'] == 'on' and moveval[compnum, 1] >= self.options['movethresh']  and moveval[compnum, 0] <= -self.options['movethresh'] :
                self.compclass[compnum] = 4

            elif self.options['muscle'] == 'on' and muscleratio[compnum] >= self.options['musclethresh']:
                self.compclass[compnum] = 5

            elif self.options['elecnoise'] == 'on' and elecnoise[compnum] >= np.abs(self.options['elecnoisethresh']):
                self.compclass[compnum] = 6

            else:
                self.compclass[compnum] = 1


        # if desired, open UI for a manual check of the components
        if self.options['compcheck'] == 'on':
            UIs.ui_select(self)

    


    def inverse_transform(self):

        self.badcomp = [i for i, comp in enumerate(self.compclass) if int(comp) != 1]
        goodcomp = [i for i in range(self.rank) if i not in self.badcomp]

        # correcting data by removing the detected artifactual components
        # dot product between the component time series and the transpose of the mixing matrix
        # mean added because data was whitened prior to ICA
        ncomps, npnts, nevents = np.shape(self.S)
        
        S_concat = np.reshape(self.S, [ncomps, -1])
        post = np.dot(S_concat[goodcomp,:].T, self.A[:,goodcomp].T)
        post += self.mean
        self.post = np.reshape(post.T, [len(self.options['chanpicks']), npnts, nevents])

        # check if satisfied with result
        if self.options['confirm'] == 'on':
            redo = UIs.ui_check(self)
        else:
            redo = False

        if redo:

            # ask if user wants to enter different settings with UI
            newsettings = UIs.ui_redo()

            if newsettings:
                self.set_options({'manualinput':'on'})
                self.fastica()

            else:
                self.set_options({'compcheck':'on'})

            self.compselect()
            self.inverse_transform()
            




    def transform_epochs_object(self):

        if self.options['remove'] == 'on':
            try:
                self.post
            except:
                self.inverse_transform()

            ch_idx = [  i for i, chan in enumerate(self.epochs.ch_names) 
                        if chan in self.options['chanpicks']]

            self.epochs._data[:, ch_idx, :] = np.moveaxis(self.post, 2, 0)
            self.transformed = True
            self.rank = np.linalg.matrix_rank(self.epochs._data)


            print('\n{} independent components removed from data.\n'.format(len(self.badcomp)))

            if len(self.badcomp) > 0:
                print('Thereof:')
                print('{} x TMS muscle artifacts'.format(sum(self.compclass==2)))
                print('{} x eye blink artifacts'.format(sum(self.compclass==3)))
                print('{} x lateral eye movement artifacts'.format(sum(self.compclass==4)))
                print('{} x persistent muscle activity'.format(sum(self.compclass==5)))
                print('{} x electrode noise \n'.format(sum(self.compclass==6)))




    def reset_orig_backend(self):

        matplotlib.use(self.orig_backend)




    def fit_select_transform(self):
        
        self.fastica()
        self.compselect()
        self.inverse_transform()
        self.transform_epochs_object()
        self.reset_orig_backend()


import mne
import numpy as np



def check_param(inst):

    options = inst.options
    epochs = inst.epochs

    # if any number of components specified, check it is less or equal to the number of channels
    if options['comps'] != -1:
        if options['comps'] > len(options['chanpicks']):
            print('\n The number of components to choose from ({}) '
                  'is more than the number of channel picks ({}).'
                  .format(options['comps'], len(options['chanpicks'])))
            print('Adjusting to maximal number of components ({}).'.format(len(options['chanpicks'])))
            options['comps'] = len(options['chanpicks'])
    else:
        options['comps'] = len(options['chanpicks'])

    # check figure inputs
    accepted_strings = ['small', 'medium', 'large']
    if options['figsize'].lower() not in accepted_strings:
        raise ValueError('Input for \'figsize\' must be either \'small\', \'medium\', or \'large\'.')

    # check the length of the subwindow specified for plotting
    if len(options['plottimex']) != 2:
        raise ValueError('Input for \'plottimex\' must be in the following format: [start, end], e.g. [-200, 500].')
    elif options['plottimex'][0] < epochs.times[0]*1000 or options['plottimex'][1] > epochs.times[-1]*1000:
        raise ValueError('Input for \'plottimex\' must be within the epoch\'s time window.')

    # check if frequency input is in the right format and has above 0 values
    if len(options['plotfreqx']) != 2:
        raise ValueError('Input for \'plotfreqx\' must be in the following format: [low, high], e.g. [1, 100].')
    elif options['plotfreqx'][0] | options['plotfreqx'][1] < 0:
        raise ValueError('Input for \'plotfreqx\' must be larger than 0.')

    if options['freqscale'].lower() not in ['raw', 'log', 'log10', 'db']:
        raise ValueError('Input for \'freqscale\' must be either \'raw\', \'log\', \'log10\' or \'db\'.')

    # check the muscle threshold input
    if options['tmsmuscle'] == 'on':
        if options['tmsmusclethresh'] < 0:
            raise ValueError('Input for \'tmsmusclethresh\' must be greater than 0.')
        elif len(options['tmsmusclewin']) != 2:
            raise ValueError('Input for \'tmsmuscleswin\' must be in the following format: [start,end]. e.g. [11,51].')
        elif options['tmsmusclewin'][0] < epochs.times[0]*1000 or options['tmsmusclewin'][1] > epochs.times[-1]*1000:
            raise ValueError(   'Input for \'tmsmuscleswin\' is out of epoch bounds.'
                                'It needs to be between {} and {} ms.'
                                .format(epochs.times[0]*1000, epochs.times[-1]*1000))

    # check options for the fast ICA function
    if options['approach'] not in ['parallel', 'deflation']:
        raise ValueError('Input for \'approach\' must be either \'parallel\' or \'deflation\'.')
    elif options['g'] not in ['logcosh', 'exp', 'cube']:
        raise ValueError('Input for \'g\' must be either \C.')

    # check frequency scaling input
    if options['freqscale'] not in ['raw', 'log', 'log10', 'db']:
        raise ValueError('Input for \'freqscale\' needs to be either \'raw\', \'log\', \'log10\' or \'db\'.')

    # check lateral eye movement input, disable if one or both electrodes are not present.
    if options['move'] == 'on':
        eNum = [i for i, chan in enumerate(options['chanpicks']) if chan in options['moveelecs']]
    
        if len(eNum) < 2:
            options['move'] = 'off'
            Warning('One or both electrodes required for detecting lateral eye movements not present.'
                    'This function has been disabled.')

    # check eyeblink input, if at least one electrode present
    if options['blink'] == 'on':
        blinkNum = [i for i, chan in enumerate(options['chanpicks']) if chan in options['blinkelecs']]
        missing = [chan for chan in options['blinkelecs'] if chan not in options['chanpicks']]

        if len(blinkNum) == 0:
            options['blink'] = 'off'
            Warning('Blink electrodes not found. Blink detection has been disabled.')
        elif len(blinkNum) == 1:
            Warning('Electrode {} is not present in the data.'
                    'Electrode not included in blink detection.'.format(missing))
        
    # check input for persistent muscle activity
    if options['muscle'] == 'on':
        if len(options['musclefreqin']) == 0:
            options['musclefreqin'] = options['plotfreqx']
        elif len(options['musclefreqin']) != 2:
            raise ValueError('Input for \'musclefreqin\' must be in the following format: [low, high], e.g. [7, 75].')
        elif options['musclefreqin'][0] < options['plotfreqx'][0] or options['musclefreqin'][1] > options['plotfreqx'][1]:
            raise ValueError('Input for \'musclefreqin\' are outside of the frequency range set by input \'plotfreqx\'.'
                            'Please adjust.')
        
        if len(options['musclefreqex']) == 0:
            pass
        elif len(options['musclefreqex']) != 2:
            raise ValueError('Input for \'musclefreqex\' must be either empty or in the following format:'
                                '[low, high], e.g. [48, 52].')
        elif options['musclefreqex'][0] < options['musclefreqin'][0] or options['musclefreqex'][1] > options['musclefreqin'][1]:
            raise ValueError('Input for \'musclefreqex\' are outside of the frequency range set by input \'musclefreqin\'.'
                                'Please adjust.')                  




def eval_param(options, kwargs):

    for opt, val in kwargs.items():

        opt = opt.lower()

        if opt in options:
            options[opt] = val
        else:
            raise ValueError('\n{} is not a recognized parameter name.'.format(opt))

    return options





def chan_visual_inspection(x, indexmode = 'exclude'):

    """
    from Github mkeute (visual_inspection).
    Allows you to visually inspect and exclude elements from an array.
    The array x typically contains summary statistics, e.g., the signal
    variance for each trial.
    """

    import matplotlib.pyplot as plt
    from matplotlib.widgets import RectangleSelector
    import numpy as np
    import matplotlib

    matplotlib.use('Qt5Agg')

    x = np.array(x)
    x = x.flatten()
    nanix = np.zeros(len(x))

    def line_select_callback(eclick, erelease):
        """
        Callback for line selection.
    
        *eclick* and *erelease* are the press and release events.
        """
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        
        print("(%3.2f, %3.2f) --> (%3.2f, %3.2f)" % (x1, y1, x2, y2))
        print(" The button you used were: %s %s" % (eclick.button, erelease.button))
    

    fig, current_ax = plt.subplots()                 # make a new plotting range
    # plt.plot(np.arange(len(x)), x, lw=1, c='b', alpha=.7)  # plot something
    current_ax.plot(x, 'b.', alpha=.7)  # plot something

    print("\n      click  -->  release")

    # drawtype is 'box' or 'line' or 'none'
    RS = RectangleSelector(current_ax, line_select_callback,
                                    drawtype='box', useblit=True,
                                    button=[1],  # don't use middle button
                                    minspanx=5, minspany=5,
                                    spancoords='pixels',
                                    interactive=True)
    RSinv = RectangleSelector(current_ax, line_select_callback,
                                drawtype='box', useblit=True,
                                button=[3],  # don't use middle button
                                minspanx=5, minspany=5,
                                spancoords='pixels',
                                interactive=True)
    plt.connect('key_press_event', (RS, RSinv))

    while plt.fignum_exists(1):
        plt.cla()
        current_ax.set_ylim([np.min(x[np.where(nanix == 0)[0]]), 1.1*np.max(x[np.where(nanix == 0)[0]])])
        current_ax.plot(x, 'b.', alpha=.7)  # plot something
        if np.sum(nanix) > 0:
            current_ax.plot(np.squeeze(np.where(nanix == 1)), x[np.where(nanix == 1)], 'w.', alpha=.7)  # plot something

        fig.show()
        plt.pause(.1)
        if plt.fignum_exists(1):
            plt.waitforbuttonpress(timeout = 2)
            
            if (RS.geometry[1][1] > 1):
                exclix = np.where((x > min(RS.geometry[0])) & (x < max(RS.geometry[0])))[0]
                exclix = exclix[np.where((exclix > min(RS.geometry[1])) & (exclix < max(RS.geometry[1])))]
                nanix[exclix] = 1
            if (RSinv.geometry[1][1] > 1):
                exclix = np.where((x > min(RSinv.geometry[0])) & (x < max(RSinv.geometry[0])))[0]
                exclix = exclix[np.where((exclix > min(RSinv.geometry[1])) & (exclix < max(RSinv.geometry[1])))]
                nanix[exclix] = 0
            if not plt.fignum_exists(1):
                break
            else:
                plt.pause(.1)
        else:
            plt.pause(.1)
            break
    if indexmode == 'exclude':
        return np.where(nanix == 1)[0]
    elif indexmode == 'keep':
        return np.where(nanix == 0)[0]
    else:
    	raise ValueError




def MNE_raw_format(eegdata, ch_names, sfreq):

    """
    Helps get EEG data into MNE raw object format by using some default values.
    Channel types all defined as EEG and default 1005 montage used.

    Args:
        eegdata (numpy array): EEG data array, timepoints*channels
        ch_names (list): list of strings with channel names
        sfreq (float, int): sampling frequency

    Returns:
        raw : instance of MNE raw
    """

    ch_types = ['eeg']*len(ch_names)
    
    info = mne.create_info(ch_names=ch_names, 
                           sfreq=sfreq, 
                           ch_types=ch_types,
                           verbose=0)

    raw = mne.io.RawArray(np.transpose(eegdata), info, verbose=0)

    raw.set_montage(mne.channels.make_standard_montage('standard_1005'), verbose=0)

    return raw




def cubic_interpolation(epochs, win):

    from scipy import interpolate

    x = epochs.times

    # convert window to indices in epochs; +1 because it should include the last timepoint
    idx1 = np.argmin(np.abs(x - win[0]*0.001))
    idx2 = np.argmin(np.abs(x - win[1]*0.001)) +1

    # delete timepoints that should be interpolated
    x = np.delete(x, np.s_[idx1:idx2], 0)

    for i, epoch in enumerate(epochs):

        y = np.delete(epoch, np.s_[idx1:idx2], -1)
        p = interpolate.interp1d(x, y, kind='cubic')

        # get the interpolation values for the timepoints of interest
        interp_values = p.__call__(epochs.times[idx1:idx2])

        # for each epoch replace timepoints in epochs object
        epochs._data[i, :, idx1:idx2] = interp_values

    return epochs



from setuptools import setup
setup(
    name='TMSRepair',
    version='0.0.10',
    description='MNE extension for repairing TMS-evoked artifacts in EEG data',
    long_description='MNE extension for repairing TMS-evoked artifacts in epochs-objects with Independent Component Analysis',
    author='Mara Wolter',
    author_email='mara.k.wolter@gmail.com',
    url='https://github.com/marakw/TMSRepair.git',
    download_url='https://github.com/marakw/TMSRepair.git',
    license='MIT',
    packages=['TMSRepair'],
    install_requires=[
        'numpy',
        'matplotlib',
        'mne',
        'scipy',
        'sklearn'
        ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
        ]
    )
